package com.lzmispt.data.resource.service.datasync.impl.statistics;

package .tmp;

import com.alibaba.fastjson.JSON;
import com.lzmispt.core.dao.database.impl.BaseDaoImpl;
import com.lzmispt.data.center.entity.vo.resource.datasource.DataSourceVO;
import com.lzmispt.data.resource.client.DataStandardClient;
import com.lzmispt.data.resource.dao.datasync.IDataSyncDao;
import com.lzmispt.data.resource.entity.dto.datasync.FlinkTaskInfoDTO;
import com.lzmispt.data.resource.entity.dto.resanddata.ResAndDataEntityDTO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncDirtyDataPO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncLogPO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncTaskPO;
import com.lzmispt.data.resource.entity.po.datasync.DataSyncTaskTablesPO;
import com.lzmispt.data.resource.entity.vo.datasync.DirtyDataRerunInfoVO;
import com.lzmispt.data.resource.entity.vo.datasync.ReadWriteStatistics;
import com.lzmispt.data.resource.entity.vo.datasync.TaskReadWriteStatistics;
import com.lzmispt.data.resource.manager.metadata.MetadataManager;
import com.lzmispt.data.resource.service.datasync.impl.DataSyncThreadPool;
import com.lzmispt.data.resource.service.datasync.impl.flink.FlinkApi;
import com.lzmispt.dcbase.dcconst.DCConst;
import com.lzmispt.flink.entity.statistics.DirtyInfo;
import com.lzmispt.flink.entity.statistics.SyncRecord;
import com.lzmispt.flink.statistics.StatisticsRecorder;
import com.lzmispt.framework.utils.JedisUtil;
import com.lzmispt.framework.utils.WorkerId;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

@Service
public class DataSyncStatisticsHandler {
  @Autowired
  JedisUtil jedisUtil;
  
  @Autowired
  BaseDaoImpl baseDao;
  
  @Autowired
  IDataSyncDao dataSyncDao;
  
  @Autowired
  @Qualifier("leadingDataLogMongo")
  MongoTemplate mongoTemplate;
  
  @Autowired
  private DataStandardClient standardClient;
  
  private static final String DATA_SYNC_STATISTICS_KEY = "DATA_SYNC_STATISTICS:";
  
  public static final String TABLE_STATISTICS_INIT = JSON.toJSONString(new ReadWriteStatistics());
  
  public static final String TASK_STATISTICS_INIT = JSON.toJSONString(new TaskReadWriteStatistics());
  
  public TaskReadWriteStatistics getTaskReadWriteStatistics(Long taskId) {
    String redisKey = getStatisticsRedisKey(taskId);
    TaskReadWriteStatistics taskStatistics = (TaskReadWriteStatistics)this.jedisUtil.getObject(redisKey, TaskReadWriteStatistics.class);
    if (taskStatistics != null)
      return taskStatistics; 
    DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
    if (task == null)
      return null; 
    TaskReadWriteStatistics statistics = (TaskReadWriteStatistics)JSON.parseObject(task.getReadWriteStatistics(), TaskReadWriteStatistics.class);
    if (statistics == null)
      statistics = new TaskReadWriteStatistics(); 
    this.jedisUtil.addObject(redisKey, statistics, 0);
    return statistics;
  }
  
  public List<DirtyDataRerunInfoVO> getDirtDataRerunList(Long taskId, Integer res, String platCode, String tokenId) {
    List<DataSyncTaskTablesPO> tables = this.dataSyncDao.selectSyncTaskDataBaseNameAndDataTableByTaskId(taskId);
    return (List<DirtyDataRerunInfoVO>)tables.stream()
      .filter(item -> (!StringUtils.isEmpty(item.getDataSyncTaskInfo()) && !StringUtils.isEmpty(item.getReadWriteStatistics())))
      .map(item -> {
          DirtyDataRerunInfoVO result = new DirtyDataRerunInfoVO();
          result.setTaskId(item.getDataSyncTaskId());
          result.setTableId(item.getId());
          result.setTargetTableDirectoryId(item.getTargetTableDirectoryId());
          result.setSourceDbName(item.getSourceDatabaseName());
          FlinkTaskInfoDTO flinkTaskInfoDTO = (FlinkTaskInfoDTO)JSON.parseObject(item.getDataSyncTaskInfo(), FlinkTaskInfoDTO.class);
          result.setCurrentStage(flinkTaskInfoDTO.getCurrent_sync_stage());
          ReadWriteStatistics statistics = (ReadWriteStatistics)JSON.parseObject(item.getReadWriteStatistics(), ReadWriteStatistics.class);
          result.setDirtyDataCount(Integer.valueOf((statistics.getError()).errorRecordCount));
          return result;
        }).filter(item -> (item.getDirtyDataCount().intValue() > 0))
      .peek(item -> {
          ResAndDataEntityDTO entity = MetadataManager.selectIsMasterMetadataEntityByDirectoryId(item.getTargetTableDirectoryId(), res, platCode, tokenId);
          item.setTargetTableName(entity.getEntityTableName());
        }).collect(Collectors.toList());
  }
  
  public void handleStatisticsMsgAsync(Map<String, SyncRecord> msg) {
    DataSyncThreadPool.sequence().submit(() -> handleStatisticsMsg(msg));
  }
  
  public void handleStatisticsMsg(Map<String, SyncRecord> msg) {
    Map<Long, SyncRecord> taskIncrementMap = (Map<Long, SyncRecord>)msg.entrySet().stream().collect(Collectors.toMap(item -> {
            String[] keys = StatisticsRecorder.reverseKey((String)item.getKey());
            return Long.valueOf(keys[0]);
          }Map.Entry::getValue, (prev, next) -> {
            SyncRecord merge = new SyncRecord();
            merge.setRead(prev.getRead() + next.getRead());
            merge.setInsert(prev.getInsert() + next.getInsert());
            merge.setUpdate(prev.getUpdate() + next.getUpdate());
            merge.setDelete(prev.getDelete() + next.getDelete());
            merge.setDirty(new ArrayList(prev.getDirty().size() + next.getDirty().size()));
            merge.getDirty().addAll(prev.getDirty());
            merge.getDirty().addAll(next.getDirty());
            return merge;
          }));
    taskIncrementMap.forEach(this::handleTaskStatistics);
    msg.forEach(this::handleTablesStatistics);
  }
  
  public void subTaskStatistics(Long taskId, ReadWriteStatistics totalSub) {
    DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
    TaskReadWriteStatistics taskStatistics = (TaskReadWriteStatistics)JSON.parseObject(task.getReadWriteStatistics(), TaskReadWriteStatistics.class);
    subAndRecalculate(taskStatistics, totalSub);
    DataSyncTaskPO taskUpdate = new DataSyncTaskPO();
    taskUpdate.setId(task.getId());
    taskUpdate.setReadWriteStatistics(JSON.toJSONString(taskStatistics));
    this.baseDao.updateSelective(taskUpdate);
    String statisticsRedisKey = getStatisticsRedisKey(taskId);
    this.jedisUtil.addObject(statisticsRedisKey, taskStatistics, 0);
  }
  
  public void handleTablesStatistics(String keyStr, SyncRecord increment) {
    String[] keys = StatisticsRecorder.reverseKey(keyStr);
    Long taskId = Long.valueOf(keys[0]);
    Long mappingId = Long.valueOf(keys[1]);
    DataSyncTaskTablesPO mapping = this.dataSyncDao.selectDataSyncTaskTableById(mappingId);
    if (mapping != null) {
      ReadWriteStatistics statistics = (ReadWriteStatistics)JSON.parseObject(mapping.getReadWriteStatistics(), ReadWriteStatistics.class);
      if (statistics == null)
        statistics = new ReadWriteStatistics(); 
      incrementStatistics(statistics, increment);
      DataSyncTaskTablesPO update = new DataSyncTaskTablesPO();
      update.setId(mappingId);
      update.setReadWriteStatistics(JSON.toJSONString(statistics));
      this.baseDao.updateSelective(update);
      for (DirtyInfo dirtyInfo : increment.getDirty()) {
        DataSyncDirtyDataPO dirtyData = new DataSyncDirtyDataPO();
        dirtyData.setId((Long)WorkerId.getNextId(Long.class));
        dirtyData.setDataSyncTaskId(taskId);
        dirtyData.setDataSyncTableId(mappingId);
        dirtyData.setDataSyncDirtyDataReason(dirtyInfo.getReason());
        dirtyData.setSql(dirtyInfo.getSql());
        dirtyData.setParameters(dirtyInfo.getParam());
        dirtyData.setUserIdRecord(mapping.getUserIdRecord());
        dirtyData.setOrgIdRecord(mapping.getOrgIdRecord());
        dirtyData.setCreateTime(new Date());
        dirtyData.setUpdateTime(new Date());
        dirtyData.setIsDelete(DCConst.FALSE);
        this.mongoTemplate.insert(dirtyData, "dc_data_sync_dirty_data");
      } 
    } 
  }
  
  public void handleTaskStatistics(Long taskId, SyncRecord increment) {
    DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
    if (task == null)
      return; 
    TaskReadWriteStatistics taskStatistics = (TaskReadWriteStatistics)JSON.parseObject(task.getReadWriteStatistics(), TaskReadWriteStatistics.class);
    if (taskStatistics == null)
      taskStatistics = new TaskReadWriteStatistics(); 
    incrementTaskStatistics(taskStatistics, increment);
    DataSyncTaskPO update = new DataSyncTaskPO();
    update.setId(taskId);
    update.setUpdateTime(new Date());
    update.setReadWriteStatistics(JSON.toJSONString(taskStatistics));
    this.baseDao.updateSelective(update);
    String statisticsRedisKey = getStatisticsRedisKey(taskId);
    this.jedisUtil.addObject(statisticsRedisKey, taskStatistics, 0);
    if (taskStatistics.getErrorRecordCount() >= task.getDirtyDataThresholdValue().intValue()) {
      terminateDataSyncTask(taskId);
      recordExceedDirtyDataThresholdLog(taskId, task.getDirtyDataThresholdValue());
    } 
  }
  
  public void incrementTaskStatistics(TaskReadWriteStatistics taskStatistics, SyncRecord increment) {
    incrementStatistics((ReadWriteStatistics)taskStatistics, increment);
    taskStatistics.setReadRecordCount((taskStatistics.getInput()).readRecordCount);
    taskStatistics.setToSyncRecordCount((taskStatistics.getToSync()).toSyncRecordCount);
    taskStatistics.setWriteRecordCount((taskStatistics.getOutput()).insertRecordCount + (taskStatistics.getOutput()).updateRecordCount + (taskStatistics.getOutput()).deleteRecordCount);
    taskStatistics.setErrorRecordCount((taskStatistics.getError()).errorRecordCount);
  }
  
  public void incrementStatistics(ReadWriteStatistics statistics, SyncRecord increment) {
    ReadWriteStatistics.Input input = statistics.getInput();
    input.readRecordCount += increment.getRead();
    ReadWriteStatistics.Output output = statistics.getOutput();
    output.insertRecordCount += increment.getInsert();
    output.updateRecordCount += increment.getUpdate();
    output.deleteRecordCount += increment.getDelete();
    ReadWriteStatistics.Error error = statistics.getError();
    error.errorRecordCount += increment.getDirty().size();
    ReadWriteStatistics.ToSync toSync = statistics.getToSync();
    toSync.toSyncRecordCount = input.readRecordCount - output.insertRecordCount - output.updateRecordCount - output.deleteRecordCount - error.errorRecordCount;
  }
  
  public void terminateDataSyncTask(Long taskId) {
    List<DataSyncTaskTablesPO> tables = this.dataSyncDao.selectSyncTaskDataBaseNameAndDataTableByTaskId(taskId);
    DataSyncTaskPO task = this.dataSyncDao.selectDataSyncTaskById(taskId);
    DataSourceVO dataSource = (DataSourceVO)this.standardClient.getPlaintextDataSourceById(task.getSourceDataSourceId(), "Collection", "0").getObj();
    String dbType = dataSource.getType();
    if (DCConst.DataSyncTaskRunStatus.IN_PROGRESS.equals(task.getRunStatus())) {
      DataSyncTaskPO update = new DataSyncTaskPO();
      update.setId(taskId);
      update.setRunStatus(DCConst.DataSyncTaskRunStatus.TERMINATION);
      this.baseDao.updateSelective(update);
      List<String> flinkJobIds = (List<String>)tables.stream().map(item -> {
            FlinkTaskInfoDTO flinkInfo = (FlinkTaskInfoDTO)JSON.parseObject(item.getDataSyncTaskInfo(), FlinkTaskInfoDTO.class);
            return (flinkInfo == null || DCConst.DataSyncJobType.RERUN_DIRTY_DATA.equals(Integer.valueOf(flinkInfo.getJob_type()))) ? null : flinkInfo.getCurrent_job_id();
          }).filter(Objects::nonNull).distinct().collect(Collectors.toList());
      flinkJobIds.forEach(item -> FlinkApi.stopByDbType(item, dbType));
    } 
  }
  
  public static String getStatisticsRedisKey(Long taskId) {
    return "DATA_SYNC_STATISTICS:" + taskId;
  }
  
  public static void plus(ReadWriteStatistics source, ReadWriteStatistics plus) {
    (source.getInput()).readRecordCount += (plus.getInput()).readRecordCount;
    (source.getOutput()).updateRecordCount += (plus.getOutput()).updateRecordCount;
    (source.getOutput()).insertRecordCount += (plus.getOutput()).insertRecordCount;
    (source.getOutput()).deleteRecordCount += (plus.getOutput()).deleteRecordCount;
    (source.getError()).errorRecordCount += (plus.getError()).errorRecordCount;
  }
  
  public static void subAndRecalculate(TaskReadWriteStatistics source, ReadWriteStatistics sub) {
    ReadWriteStatistics.Input input = source.getInput();
    ReadWriteStatistics.Output output = source.getOutput();
    ReadWriteStatistics.Error error = source.getError();
    input.readRecordCount -= (sub.getInput()).readRecordCount;
    output.updateRecordCount -= (sub.getOutput()).updateRecordCount;
    output.insertRecordCount -= (sub.getOutput()).insertRecordCount;
    output.deleteRecordCount -= (sub.getOutput()).deleteRecordCount;
    error.errorRecordCount -= (sub.getError()).errorRecordCount;
    (source.getToSync()).toSyncRecordCount = input.readRecordCount - output.insertRecordCount - output.updateRecordCount - output.deleteRecordCount - error.errorRecordCount;
    source.setReadRecordCount(input.readRecordCount);
    source.setToSyncRecordCount((source.getToSync()).toSyncRecordCount);
    source.setWriteRecordCount(output.insertRecordCount + output.updateRecordCount + output.deleteRecordCount);
    source.setErrorRecordCount(error.errorRecordCount);
  }
  
  public void recordExceedDirtyDataThresholdLog(Long taskId, Integer dirtyDataThresholdValue) {
    DataSyncLogPO insert = new DataSyncLogPO();
    insert.setId((Long)WorkerId.getNextId(Long.class));
    insert.setTaskId(taskId);
    insert.setMsg((new SimpleDateFormat("yyyy-MM-dd HH:mm:ss")).format(new Date()) + " ERROR 已达到脏数据阈值【" + dirtyDataThresholdValue + "】，中止任务");
    insert.setLevel("ERROR");
    Date now = new Date();
    insert.setCreateTime(now);
    insert.setUpdateTime(now);
    insert.setIsDelete(DCConst.FALSE);
    this.mongoTemplate.insert(insert, "dc_data_sync_log");
  }
}
